@@ -4,6 +4,8 @@ use minidom::Element;
 use std::error::Error;
 use std::path::Path;
 
+const VERSION: &str = env!("CARGO_PKG_VERSION");
+
 use lib::args;
 
 fn main() {
@@ -21,6 +23,10 @@ fn main() {
         args::print_usage(&args[0], &opts);
         return;
     }
+    if parsed_args.print_version {
+        println!("sgf-render {}", VERSION);
+        return;
+    }
 
     let input = match read_input(parsed_args.infile) {
         Ok(goban) => goban,
